@@ -2,22 +2,58 @@ import pandas as pd
 import numpy as np
 import unicodedata
 import re
-from Levenshtein import distance
+from Levenshtein import distance, ratio
 import itertools
 from sklearn.neural_network import MLPClassifier
+from copy import deepcopy
+from .briefrecord import BriefRecFactory
 
 from typing import Callable, List, Tuple, Dict, Any, Optional
 
 
+def convert_to_numeric(txt: str) -> str:
+    """convert_to_numeric(str) -> str
+    Convert a text number to a numeric string.
+
+    :param txt: string to convert
+
+    :return: numeric string
+    """
+    convert_table = {'first ': '1 ',
+                     'second ': '2 ',
+                     'third ': '3 ',
+                     'fourth ': '4 ',
+                     'fifth ': '5 ',
+                     'première ': '1 ',
+                     'deuxième ': '2 ',
+                     'seconde ': '2 ',
+                     'sec. ': '2 ',
+                     'troisième ': '3 ',
+                     'quatrième ': '4 ',
+                     'erste ': '1 ',
+                     'zweite ': '2 ',
+                     'dritte ': '3 ',
+                     'vierte ': '4 '}
+    txt = txt.lower()
+    for k in convert_table:
+        txt = txt.replace(k, convert_table[k])
+
+    return txt
+
+
 def handling_missing_values(fn: Callable) -> Callable:
     """Decorator to handle missing values."""
     def wrapper(val1, val2):
 
-        if val1 is None or val2 is None:
-            return np.nan
+        if val1 is None and val2 is None:
+            return 1
+        elif val1 is None or val2 is None:
+            return 0
         if type(val1) is list or type(val2) is list:
-            if len(val1) == 0 or len(val2) == 0:
-                return np.nan
+            if len(val1) == 0 and len(val2) == 0:
+                return 1
+            elif len(val1) == 0 or len(val2) == 0:
+                return 0
 
         return fn(val1, val2)
 
@@ -66,7 +102,7 @@ def evaluate_identifiers(ids1: str, ids2: str) -> float:
         score = len(set.intersection(ids1, ids2)) / len(set.union(ids1, ids2))
         return score ** .05 if score > 0 else 0
     else:
-        return np.nan
+        return 0
 
 
 @handling_missing_values
@@ -99,13 +135,64 @@ def evaluate_sysnums(ids1: str, ids2: str) -> float:
     prefixes_ids2 = set([get_prefix(recid) for recid in ids2 if get_prefix(recid) is not None])
     if len(set.intersection(prefixes_ids1, prefixes_ids2)) == 0:
         # No common prefix: difference means nothing
-        return np.nan
+        return 0
 
     elif len(set.union(ids1, ids2)) > 0:
         score = len(set.intersection(ids1, ids2)) / len(set.union(ids1, ids2))
         return score ** .05 if score > 0 else 0
     else:
-        return np.nan
+        return 0
+
+
+@handling_missing_values
+def evaluate_language(lang1: str, lang2: str) -> float:
+    """evaluate_language(lang1: str, lang2: str) -> float
+    Return the result of the evaluation of similarity of two languages.
+
+    :param lang1: language to compare
+    :param lang2: language to compare
+
+    :return: similarity score between two languages as float
+    """
+    if lang1 == lang2:
+        return 1
+    else:
+        return 0
+
+
+def evaluate_is_analytical(format1: str, format2: str) -> float:
+    """evaluate_is_analytical(format1: str, format2: str) -> float
+    Check if records are analytical records
+
+    :param format1: format to compare
+    :param format2: format to compare
+
+    :return: 0 if no analytical, 1 if both analytical, 0.5 otherwise.
+    """
+    if format1[1] == 'a' and format2[1] == 'a':
+        return 1
+    elif format1[1] == 'a' or format2[1] == 'a':
+        return 0.5
+    else:
+        return 0
+
+
+def evaluate_is_series(format1: str, format2: str) -> Optional[float]:
+    """evaluate_is_series(format1: str, format2: str) -> Optional[float]
+    Check if records are series
+
+    :param format1: format to compare
+    :param format2: format to compare
+
+    :return: nan if no series, 1 if both series, 0.5 otherwise.
+    """
+
+    if format1[1] == 's' and format2[1] == 's':
+        return 1
+    elif format1[1] == 's' or format2[1] == 's':
+        return 0.5
+    else:
+        return 0
 
 
 @handling_missing_values
@@ -290,18 +377,28 @@ def evaluate_texts(text1: str, text2: str) -> float:
     """
     if len(text1) < len(text2):
         text1, text2 = (text2, text1)
-    coef = len(text1) - len(text2) + 1
-    text1_ascii = get_ascii(text1)
-    text2_ascii = get_ascii(text2)
 
-    score = distance(text1, text2, weights=(1, 0, 1)) * coef
-    score += distance(text1, text2, weights=(0, 1, 0))
-    score += distance(text1_ascii, text2_ascii, weights=(1, 0, 1)) * coef * 4
-    score += distance(text1_ascii, text2_ascii, weights=(0, 1, 0)) * 4
+    t_list1 = re.findall(r'\b\w+\b', text1)
+    t_list2 = re.findall(r'\b\w+\b', text2)
+    if len(t_list1) < len(t_list2):
+        t_list1, t_list2 = (t_list2, t_list1)
 
-    sum_max = (coef * len(text2) + coef - 1) * 5
+    diff = len(t_list1) - len(t_list2)
+    coef = 1 / diff**0.05 - 0.15 if diff > 0 else 1
 
-    return 1 - score / sum_max
+    score_raw = 0
+    for pos in range(len(t_list1) - len(t_list2) + 1):
+        temp_score = np.mean([ratio(t_list1[i + pos], t_list2[i]) for i in range(len(t_list2))])
+        if temp_score > score_raw:
+            score_raw = temp_score
+
+    score_ascii = 0
+    for pos in range(len(t_list1) - len(t_list2) + 1):
+        temp_score = np.mean([ratio(get_ascii(t_list1[i + pos]), get_ascii(t_list2[i])) for i in range(len(t_list2))])
+        if temp_score > score_ascii:
+            score_ascii = temp_score
+
+    return (score_raw + score_ascii * 4) / 5 * coef
 
 
 @handling_missing_values
@@ -315,7 +412,137 @@ def evaluate_format(format1: str, format2: str) -> float:
     :param format2: format to compare
 
     :return: similarity score between two formats as float"""
-    return float(format1 == format2)
+    format1 = format1.split('/')
+    format2 = format2.split('/')
+    leader1 = format1[0].strip()
+    leader2 = format2[0].strip()
+
+    # Compare leader => 0.4 max
+    score = 0.4 if leader1 == leader2 else 0
+
+    # Compare fields 33X => 0.6 max for the 3 fields
+    f33x_1 = format1[1].strip().split(';')
+    f33x_2 = format2[1].strip().split(';')
+    for i in range(len(f33x_1)):
+        if len(f33x_2) > i:
+            if f33x_1[i] == f33x_2[i]:
+                score += 0.2
+            elif f33x_1[i].strip() == '' or f33x_2[i].strip() == '':
+                score += 0.1
+
+    return score
+
+
+@handling_missing_values
+def evaluate_parents(parent1: Dict, parent2: Dict) -> float:
+    """evaluate_parents(parent1: Dict, parent2: Dict) -> float
+    Evaluate similarity based on the link to the parent
+
+    Keys of the parent dictionary:
+    - title: title of the parent
+    - issn: content of $x
+    - isbn: content of $z
+    - number: content of $g no:<content>
+    - year: content of $g yr:<content> or first 4 digits numbers in a $g
+    - parts: longest list of numbers in a $g
+
+    :param parent1: dictionary with parent information
+    :param parent2: dictionary with parent information
+
+    :return: similarity score between two parents
+    """
+
+    score_title = 0
+    score_identifiers = None
+    score_year = None
+    score_no = None
+    score_parts = None
+
+    if 'title' in parent1 and 'title' in parent2:
+        score_title = evaluate_texts(parent1['title'], parent2['title'])
+
+    if 'issn' in parent1 and 'issn' in parent2:
+        score_identifiers = evaluate_identifiers([parent1['issn']], [parent2['issn']])
+    elif 'isbn' in parent1 and 'isbn' in parent2:
+        score_identifiers = evaluate_identifiers([parent1['isbn']], [parent2['isbn']])
+
+    if 'number' in parent1 and 'number' in parent2:
+        score_no = int(BriefRecFactory.normalize_extent(parent1['number']) ==
+                       BriefRecFactory.normalize_extent(parent2['number']))
+
+    if 'year' in parent1 and 'year' in parent2:
+        score_year = int(parent1['year'] == parent2['year'])
+
+    for p in [parent1, parent2]:
+
+        # Create parts field if not present => used to compare two records with parts fields
+        if 'parts' not in p:
+            parts = []
+            if 'number' in p:
+                parts += BriefRecFactory.normalize_extent(p['number'])
+            if 'year' in p:
+                parts.append(p['year'])
+            if len(parts) > 0:
+                p['parts'] = parts
+
+    parent1 = deepcopy(parent1)
+    parent2 = deepcopy(parent2)
+    if 'parts' in parent1 and 'parts' in parent2:
+        initial_nb = sum([len(parent1['parts']), len(parent2['parts'])])
+        to_delete = []
+        for e in parent1['parts']:
+            if e in parent2['parts']:
+                to_delete.append(e)
+                parent2['parts'].remove(e)
+        for e in to_delete:
+            parent1['parts'].remove(e)
+
+        to_delete = []
+        for e in parent2['parts']:
+            if e in parent1['parts']:
+                to_delete.append(e)
+                parent1['parts'].remove(e)
+        for e in to_delete:
+            parent2['parts'].remove(e)
+
+        final_nb = sum([len(parent1['parts']), len(parent2['parts'])])
+        score_parts = 1 - final_nb / initial_nb
+
+    elif 'parts' in parent1 or 'parts' in parent2:
+        # Case if part information is only in one record available
+        score_parts = 0
+
+    return score_title * np.mean([s for s in [score_title, score_no, score_year, score_identifiers, score_parts]
+                                  if s is not None])
+
+
+@handling_missing_values
+def evaluate_editions(texts1: List[str], texts2: List[str]) -> float:
+    """evaluate_editions(texts1: List[str], texts2: List[str]) -> float
+    Return the result of the evaluation of similarity of two editions. If numbers are available,
+    these are preferred to texts
+
+    :param texts1: list of editions to compare
+    :param texts2: list of editions to compare
+
+    :return: similarity score between two editions as float
+    """
+    score_txt = evaluate_lists_texts(texts1, texts2)
+
+    nb_list_1 = []
+    nb_list_2 = []
+    for txt in texts1:
+        nb_list_1 += [int(f) for f in re.findall(r'\d+', convert_to_numeric(txt))]
+    for txt in texts2:
+        nb_list_2 += [int(f) for f in re.findall(r'\d+', convert_to_numeric(txt))]
+
+    nb_list_1 = set(nb_list_1)
+    nb_list_2 = set(nb_list_2)
+    if len(nb_list_1) > 0 and len(nb_list_2) > 0:
+        score_nb = len(set.intersection(nb_list_1, nb_list_2)) / max(len(nb_list_1), len(nb_list_2))
+        return (score_txt + score_nb * 9) / 10
+    else:
+        return score_txt
 
 
 @handling_missing_values
@@ -352,9 +579,10 @@ def evaluate_similarity(bib1: Dict[str, Any], bib2: Dict[str, Any]) -> Dict[str,
         'format': evaluate_format(bib1['format'], bib2['format']),
         'title': evaluate_texts(bib1['title'], bib2['title']),
         'short_title': evaluate_texts(bib1['short_title'], bib2['short_title']),
-        'editions': evaluate_lists_texts(bib1['editions'], bib2['editions']),
+        'editions': evaluate_editions(bib1['editions'], bib2['editions']),
         'creators': evaluate_lists_names(bib1['creators'], bib2['creators']),
         'corp_creators': evaluate_lists_names(bib1['corp_creators'], bib2['corp_creators']),
+        'language': evaluate_language(bib1['language'], bib2['language']),
         'date_1': evaluate_year(bib1['date_1'], bib2['date_1']),
         'date_2': evaluate_year(bib1['date_2'], bib2['date_2']),
         'publishers': evaluate_lists_texts(bib1['publishers'], bib2['publishers']),
@@ -363,8 +591,11 @@ def evaluate_similarity(bib1: Dict[str, Any], bib2: Dict[str, Any]) -> Dict[str,
         'isbns': evaluate_identifiers(bib1['isbns'], bib2['isbns']),
         'issns': evaluate_identifiers(bib1['issns'], bib2['issns']),
         'other_std_num': evaluate_identifiers(bib1['other_std_num'], bib2['other_std_num']),
+        'parent': evaluate_parents(bib1['parent'], bib2['parent']),
         'sysnums': evaluate_sysnums(bib1['sysnums'], bib2['sysnums']),
-        'same_fields_existing': evaluate_completeness(bib1, bib2)
+        'same_fields_existing': evaluate_completeness(bib1, bib2),
+        'are_analytical': evaluate_is_analytical(bib1['format'], bib2['format']),
+        'are_series': evaluate_is_series(bib1['format'], bib2['format']),
     }
     return results
 
@@ -397,6 +628,7 @@ def get_similarity_score(bib1: Dict[str, Any],
                    'editions',
                    'creators',
                    'corp_creators',
+                   'language',
                    'date_1',
                    'date_2',
                    'publishers',
@@ -405,8 +637,11 @@ def get_similarity_score(bib1: Dict[str, Any],
                    'isbns',
                    'issns',
                    'other_std_num',
+                   'parent',
                    'sysnums',
-                   'same_fields_existing']
+                   'same_fields_existing',
+                   'are_analytical',
+                   'are_series']
 
         # Prepare DataFrame to save similarity evaluations
         df = pd.DataFrame(columns=columns)
@@ -417,4 +652,5 @@ def get_similarity_score(bib1: Dict[str, Any],
         # 0.5 is the threshold to determine if two records are similar or not
         return clf.predict_proba(df)[0][1]
     else:
-        return np.mean([results[k] for k in results if pd.isna(results[k]) is False])
+        return np.mean([results[k] for k in results if pd.isna(results[k]) is False and k not in ['are_analytical',
+                                                                                                  'are_series']])
